@@ -1,12 +1,13 @@
 #!/usr/bin/env python
 
-import collections
 import math
 import operator
 import warnings
+from copy import deepcopy
 
 import geopandas as gpd
 import libpysal
+import networkx as nx
 import numpy as np
 import pandas as pd
 import shapely
@@ -14,11 +15,13 @@ from packaging.version import Version
 from scipy.signal import find_peaks
 from scipy.stats import gaussian_kde
 from shapely.geometry import LineString, Point
-from shapely.ops import linemerge, polygonize
+from shapely.ops import linemerge, polygonize, split
 from tqdm.auto import tqdm
 
 from .coins import COINS
+from .graph import node_degree
 from .shape import CircularCompactness
+from .utils import gdf_to_nx, nx_to_gdf
 
 __all__ = [
     "preprocess",
@@ -27,6 +30,7 @@ __all__ = [
     "close_gaps",
     "extend_lines",
     "roundabout_simplification",
+    "consolidate_intersections",
     "FaceArtifacts",
 ]
 
@@ -84,7 +88,9 @@ def preprocess(
         print("Loop", loop + 1, f"out of {loops}.") if verbose else None
         blg.reset_index(inplace=True, drop=True)
         blg["mm_uid"] = range(len(blg))
-        sw = libpysal.weights.contiguity.Rook.from_dataframe(blg, silence_warnings=True)
+        sw = libpysal.weights.contiguity.Rook.from_dataframe(
+            blg, silence_warnings=True, use_index=False
+        )
         blg["neighbors"] = sw.neighbors.values()
         blg["n_count"] = blg.apply(lambda row: len(row.neighbors), axis=1)
         blg["circu"] = CircularCompactness(blg).series
@@ -148,10 +154,10 @@ def preprocess(
                     if not subset.empty:
                         geoms.append(blg[blg["mm_uid"] == j].iloc[0].geometry)
                         blg.drop(blg[blg["mm_uid"] == j].index[0], inplace=True)
-                new_geom = shapely.ops.unary_union(geoms)
-                blg.loc[
-                    blg.loc[blg["mm_uid"] == key].index[0], blg.geometry.name
-                ] = new_geom
+                new_geom = shapely.union_all(geoms)
+                blg.loc[blg.loc[blg["mm_uid"] == key].index[0], blg.geometry.name] = (
+                    new_geom
+                )
 
         blg.drop(delete, inplace=True)
     return blg[buildings.columns]
@@ -160,8 +166,8 @@ def preprocess(
 def remove_false_nodes(gdf):
     """
     Clean topology of existing LineString geometry by removal of nodes of degree 2.
-
-    Returns the original gdf if there's no node of degree 2.
+    Returns the original gdf if there's no node of degree 2. Some geometries may
+    be forced to 2D where a Z coordinate is present.
 
     Parameters
     ----------
@@ -172,7 +178,7 @@ def remove_false_nodes(gdf):
     -------
     gdf : GeoDataFrame, GeoSeries
 
-    See also
+    See Also
     --------
     momepy.extend_lines
     momepy.close_gaps
@@ -189,63 +195,97 @@ def remove_false_nodes(gdf):
         df = gpd.GeoSeries(gdf)
 
     # extract array of coordinates and number per geometry
-    coords = shapely.get_coordinates(geom)
-    indices = shapely.get_num_coordinates(geom)
+    start_points = shapely.get_point(geom, 0)
+    end_points = shapely.get_point(geom, -1)
 
-    # generate a list of start and end coordinates and create point geometries
-    edges = [0]
-    i = 0
-    for ind in indices:
-        ix = i + ind
-        edges.append(ix - 1)
-        edges.append(ix)
-        i = ix
-    edges = edges[:-1]
-    points = shapely.points(np.unique(coords[edges], axis=0))
+    points = shapely.points(
+        np.unique(
+            shapely.get_coordinates(np.concatenate([start_points, end_points])), axis=0
+        )
+    )
 
     # query LineString geometry to identify points intersecting 2 geometries
     tree = shapely.STRtree(geom)
     inp, res = tree.query(points, predicate="intersects")
     unique, counts = np.unique(inp, return_counts=True)
-    merge = res[np.isin(inp, unique[counts == 2])]
+    mask = np.isin(inp, unique[counts == 2])
+    merge_res = res[mask]
+    merge_inp = inp[mask]
 
-    if len(merge):
-        # filter duplications and create a dictionary with indication of components to
-        # be merged together
-        dups = [item for item, count in collections.Counter(merge).items() if count > 1]
-        split = np.split(merge, len(merge) / 2)
-        components = {}
-        for i, a in enumerate(split):
-            if a[0] in dups or a[1] in dups:
-                if a[0] in components:
-                    i = components[a[0]]
-                elif a[1] in components:
-                    i = components[a[1]]
-            components[a[0]] = i
-            components[a[1]] = i
+    if len(merge_res):
+        g = nx.Graph(list(zip(merge_inp * -1, merge_res, strict=True)))
+        new_geoms = []
+        for c in nx.connected_components(g):
+            valid = [ix for ix in c if ix > -1]
+            new_geoms.append(shapely.line_merge(shapely.union_all(geom[valid])))
 
-        # iterate through components and create new geometries
-        all_keys = {}
-        for k, v in components.items():
-            all_keys.setdefault(v, []).append(k)
-        new = []
-        for keys in all_keys.values():
-            new.append(shapely.line_merge(shapely.union_all(geom[keys])))
-
-        # remove incorrect geometries and append fixed versions
-        df = df.drop(merge)
-        final = gpd.GeoSeries(new).explode(ignore_index=True)
+        df = df.drop(merge_res)
+        final = gpd.GeoSeries(new_geoms, crs=df.crs).explode(ignore_index=True)
         if isinstance(gdf, gpd.GeoDataFrame):
-            return pd.concat(
+            combined = pd.concat(
                 [
                     df,
                     gpd.GeoDataFrame(
-                        {df.geometry.name: final}, geometry=df.geometry.name
+                        {df.geometry.name: final}, geometry=df.geometry.name, crs=df.crs
                     ),
                 ],
                 ignore_index=True,
             )
-        return pd.concat([df, final], ignore_index=True)
+        else:
+            combined = pd.concat([df, final], ignore_index=True)
+
+        # re-order closed loops
+        fixed_loops = []
+        fixed_index = []
+        nodes = nx_to_gdf(
+            node_degree(
+                gdf_to_nx(
+                    combined
+                    if isinstance(combined, gpd.GeoDataFrame)
+                    else combined.to_frame("geometry")
+                )
+            ),
+            lines=False,
+        )
+
+        loops = combined[combined.is_ring]
+
+        node_ix, loop_ix = loops.sindex.query(nodes.geometry, predicate="intersects")
+        for ix in np.unique(loop_ix):
+            loop_geom = loops.geometry.iloc[ix]
+            target_nodes = nodes.geometry.iloc[node_ix[loop_ix == ix]]
+            if len(target_nodes) == 2:
+                node_coords = shapely.get_coordinates(target_nodes)
+                coords = shapely.get_coordinates(loop_geom)
+                new_start = (
+                    node_coords[0]
+                    if (node_coords[0] != coords[0]).all()
+                    else node_coords[1]
+                )
+                new_start_idx = np.where(coords == new_start)[0][0]
+                rolled_coords = np.roll(coords[:-1], -new_start_idx, axis=0)
+                new_sequence = np.append(rolled_coords, rolled_coords[[0]], axis=0)
+                fixed_loops.append(shapely.LineString(new_sequence))
+                fixed_index.append(ix)
+        fixed_loops = gpd.GeoSeries(fixed_loops, crs=df.crs).explode(ignore_index=True)
+
+        if isinstance(gdf, gpd.GeoDataFrame):
+            return pd.concat(
+                [
+                    combined.drop(loops.iloc[fixed_index].index),
+                    gpd.GeoDataFrame(
+                        {df.geometry.name: fixed_loops},
+                        geometry=df.geometry.name,
+                        crs=df.crs,
+                    ),
+                ],
+                ignore_index=True,
+            )
+        else:
+            return pd.concat(
+                [combined.drop(loops.iloc[fixed_index].index), fixed_loops],
+                ignore_index=True,
+            )
 
     # if there's nothing to fix, return the original dataframe
     return gdf
@@ -753,7 +793,7 @@ def _selecting_rabs_from_poly(
         rab_adj.index.name = "index"
 
         # adding a hausdorff_distance threshold
-        rab_adj["hdist"] = 0
+        rab_adj["hdist"] = 0.0
         # TODO: (should be a way to vectorize)
         for i, group in rab_adj.groupby("index_right"):
             for g in group.itertuples():
@@ -960,62 +1000,53 @@ def roundabout_simplification(
 
     If ``include_adjacent`` is True, adjacent polygons to the actual roundabout are
     also selected for simplification if two conditions are met:
-        - the area of adjacent polygons is less than the actual roundabout
-        - adjacent polygons do not extend beyond a factor of the diameter of the actual
-        roundabout.
-        This uses hausdorff_distance algorithm.
+
+    - the area of adjacent polygons is less than the actual roundabout
+    - adjacent polygons do not extend beyond a factor of the diameter of the actual
+      roundabout. This uses hausdorff_distance algorithm.
 
     Parameters
     ----------
     edges : GeoDataFrame
         GeoDataFrame containing LineString geometry of urban network
     polys : GeoDataFrame
-        GeoDataFrame containing Polygon geometry derived from polygonyzing
-        ``edges`` GeoDataFrame.
+        GeoDataFrame containing Polygon geometry derived from polygonyzing ``edges``
+        GeoDataFrame.
     area_col : string
         Column name containing area values if ``polys`` GeoDataFrame contains such
         information. Otherwise, it will
     circom_threshold : float (default 0.7)
         Circular compactness threshold to select roundabouts from ``polys``
-        GeoDataFrame.
-        Polygons with a higher or equal threshold value will be considered for
-        simplification.
+        GeoDataFrame. Polygons with a higher or equal threshold value will be considered
+        for simplification.
     area_threshold : float (default 0.85)
         Percentile threshold value from the area of ``polys`` to leave as input
-        geometry.
-        Polygons with a higher or equal threshold will be considered as urban blocks
-        not considered
-        for simplification.
+        geometry. Polygons with a higher or equal threshold will be considered as urban
+        blocks not considered for simplification.
     include_adjacent : boolean (default True)
         Adjacent polygons to be considered also as part of the simplification.
     diameter_factor : float (default 1.5)
-        The factor to be applied to the diameter of each roundabout that determines
-        how far an adjacent polygon can stretch until it is no longer considered part
-        of the overall roundabout group. Only applyies when include_adjacent = True.
+        The factor to be applied to the diameter of each roundabout that determines how
+        far an adjacent polygon can stretch until it is no longer considered part of the
+        overall roundabout group. Only applyies when include_adjacent = True.
     center_type : string (default 'centroid')
-        Method to use for converging the incoming LineStrings.
-        Current list of options available : 'centroid', 'mean'.
-        - 'centroid': selects the centroid of the actual roundabout (ignoring adjacent
-        geometries)
-        - 'mean': calculates the mean coordinates from the points of polygons (including
-         adjacent geometries)
+        Method to use for converging the incoming LineStrings. Current list of options
+        available : 'centroid', 'mean'. - 'centroid': selects the centroid of the actual
+        roundabout (ignoring adjacent geometries) - 'mean': calculates the mean
+        coordinates from the points of polygons (including adjacent geometries)
     angle_threshold : int, float (default 0)
         The angle threshold for the COINS algorithm. Only used when multiple incoming
-        LineStrings
-        arrive at the same Point to the roundabout or to the adjacent polygons if set
-        as True.
-        eg. when two 'edges' touch the roundabout at the same point, COINS algorithm
-        will evaluate which of those
-        incoming lines should be extended according to their deflection angle.
-        Segments will only be considered a part of the same street if the deflection
-        angle
-        is above the threshold.
+        LineStrings arrive at the same Point to the roundabout or to the adjacent
+        polygons if set as True. eg. when two 'edges' touch the roundabout at the same
+        point, COINS algorithm will evaluate which of those incoming lines should be
+        extended according to their deflection angle. Segments will only be considered a
+        part of the same street if the deflection angle is above the threshold.
 
     Returns
     -------
     GeoDataFrame
-        GeoDataFrame with an updated geometry and an additional
-        column labeling modified edges.
+        GeoDataFrame with an updated geometry and an additional column labeling modified
+        edges.
     """
     if len(edges[edges.geom_type != "LineString"]) > 0:
         raise TypeError(
@@ -1041,8 +1072,459 @@ def roundabout_simplification(
     return output
 
 
+def consolidate_intersections(
+    graph,
+    tolerance=30,
+    rebuild_graph=True,
+    rebuild_edges_method="spider",
+    x_att="x",
+    y_att="y",
+    edge_from_att="from",
+    edge_to_att="to",
+):
+    """
+    Consolidate close street intersections into a single node, collapsing short edges.
+
+    If rebuild_graph is True, new edges are drawn according to ``rebuild_edges_method``
+    which is one of:
+
+    1. Extension reconstruction:
+        Edges are linearly extended from original endpoints until the new nodes. This
+        method preserves most faithfully the network geometry but can result in
+        overlapping geometry.
+    2. Spider-web reconstruction:
+        Edges are cropped within a buffer of the new endpoints and linearly extended
+        from there. This method improves upon linear reconstruction by mantaining, when
+        possible, network planarity.
+    3. Euclidean reconstruction:
+        Edges are ignored and new edges are built as straight lines between new origin
+        and new destination. This method ignores geometry, but efficiently preserves
+        adjacency.
+
+    If ``rebuild_graph`` is False, graph is returned with consolidated nodes but without
+    reconstructed edges i.e. graph is intentionally disconnected.
+
+    Graph must be configured so that
+
+    1. All nodes have attributes determining their x and y coordinates;
+    2. All edges have attributes determining their origin, destination, and geometry.
+
+    Parameters
+    ----------
+    graph : Networkx.Graph, Networkx.DiGraph, Networkx.MultiGraph, or
+        Networkx.MultiDiGraph
+    tolerance : float, default 30
+        distance in network units below which nodes will be consolidated
+    rebuild_graph : bool
+    rebuild_edges_method : str
+        'extend' or 'spider' or 'euclidean', ignored if rebuild_graph is False
+    x_att : str
+        node attribute with the valid x-coordinate
+    y_att : str
+        node attribute with the valid y-coordinate
+    edge_from_att : str
+        edge attribute with the valid origin node id
+    edge_to_att : str
+        edge attribute with the valid destination node id
+
+    Returns
+    -------
+    Networkx.MultiGraph or Networkx.MultiDiGraph
+        directionality inferred from input type
+
+    """
+    # Collect nodes and their data:
+    nodes, nodes_dict = zip(*graph.nodes(data=True), strict=False)
+    nodes_df = pd.DataFrame(nodes_dict, index=nodes)
+    graph_crs = graph.graph.get("crs")
+
+    # Create a graph without the edges above a certain length and clean it
+    #  from isolated nodes (the unsimplifiable nodes):
+    components_graph = deepcopy(graph)
+    components_graph.remove_edges_from(
+        [
+            edge
+            for edge in graph.edges(keys=True, data=True)
+            if edge[-1]["length"] > tolerance
+        ]
+    )
+    isolated_nodes_list = list(nx.isolates(components_graph))
+    components_graph.remove_nodes_from(isolated_nodes_list)
+
+    # The connected components of this graph are node clusters we must individually
+    #  simplify. We collect them in a dataframe and retrieve node properties (x, y
+    #  coords mainly) from the original graph.
+    components = nx.connected_components(components_graph)
+    components_dict = dict(enumerate(components, start=max(nodes) + 1))
+    nodes_to_merge_dict = {
+        node: cpt for cpt, nodes in components_dict.items() for node in nodes
+    }
+    new_nodes_df = pd.DataFrame.from_dict(
+        nodes_to_merge_dict, orient="index", columns=["cluster"]
+    )
+    nodes_to_merge_df = pd.concat(
+        [new_nodes_df, nodes_df[[x_att, y_att]]], axis=1, join="inner"
+    )
+
+    # The two node attributes we need for the clusters are the position of the cluster
+    #  centroids. Those are obtained by averaging the x and y columns. We also add
+    # . attribtues referring to the original node ids in every cluster:
+    cluster_centroids_df = nodes_to_merge_df.groupby("cluster").mean()
+    cluster_centroids_df["simplified"] = True
+    cluster_centroids_df["original_node_ids"] = cluster_centroids_df.index.map(
+        components_dict
+    )
+    cluster_geometries = gpd.points_from_xy(
+        cluster_centroids_df[x_att], cluster_centroids_df[y_att]
+    )
+    cluster_gdf = gpd.GeoDataFrame(
+        cluster_centroids_df, crs=graph_crs, geometry=cluster_geometries
+    )
+    cluster_nodes_list = list(cluster_gdf.to_dict("index").items())
+
+    # Create a simplified graph object:
+    simplified_graph = graph.copy()
+
+    # Rebuild edges if necessary:
+    if rebuild_graph:
+        rebuild_edges_method = rebuild_edges_method.lower()
+        simplified_graph.graph["approach"] = "primal"
+        edges_gdf = nx_to_gdf(simplified_graph, points=False, lines=True)
+        simplified_edges = _get_rebuilt_edges(
+            edges_gdf,
+            nodes_to_merge_dict,
+            cluster_gdf,
+            method=rebuild_edges_method,
+            buffer=1.5 * tolerance,
+            edge_from_att=edge_from_att,
+            edge_to_att=edge_to_att,
+        )
+
+    # Replacing the collapsed nodes with centroids and adding edges:
+    simplified_graph.remove_nodes_from(nodes_to_merge_df.index)
+    simplified_graph.add_nodes_from(cluster_nodes_list)
+    if rebuild_graph:
+        simplified_graph.add_edges_from(simplified_edges)
+
+    return simplified_graph
+
+
+def _get_rebuilt_edges(
+    edges_gdf,
+    nodes_dict,
+    cluster_gdf,
+    method="spider",
+    buffer=45,
+    edge_from_att="from",
+    edge_to_att="to",
+):
+    """
+    Update origin and destination on network edges when original endpoints were replaced
+    by a
+      consolidated node cluster. New edges are drawn according to method which is one
+      of:
+
+    1. Extension reconstruction:
+        Edges are linearly extended from original endpoints until the new nodes. This
+        method preserves most faithfully the network geometry.
+    2. Spider-web reconstruction:
+        Edges are cropped within a buffer of the new endpoints and linearly extended
+        from there. This method improves upon linear reconstruction by mantaining, when
+        possible, network planarity.
+    3. Euclidean reconstruction:
+        Edges are ignored and new edges are built as straightlines between new origin
+        and new destination. This method ignores geometry, but efficiently preserves
+        adjacency.
+
+    Parameters
+    ----------
+    edges_gdf : GeoDataFrame
+        GeoDataFrame containing LineString geometry and columns determining origin
+        and destination node ids
+    nodes_dict : dict
+        Dictionary whose keys are node ids and values are the corresponding consolidated
+        node cluster ids. Only consolidated nodes are in the dictionary.
+    cluster_gdf : GeoDataFrame
+        GeoDataFrame containing consolidated node ids.
+    method: string
+        'extension' or 'spider' or 'euclidean'
+    buffer : float
+        distance to buffer consolidated nodes in the Spider-web reconstruction
+    edge_from_att : str
+        edge attribute with the valid origin node id
+    edge_to_att : str
+        edge attribute with the valid destination node id
+
+    Returns
+    ----------
+    List
+        list of edges that should be added to the network. Edges are in the format
+        (origin_id, destination_id, data), where data is inferred from edges_gdf
+
+    """
+    # Determine what endpoints were made into clusters:
+    edges_gdf["origin_cluster"] = edges_gdf[edge_from_att].apply(
+        lambda u: nodes_dict.get(u, -1)
+    )
+    edges_gdf["destination_cluster"] = edges_gdf[edge_to_att].apply(
+        lambda v: nodes_dict.get(v, -1)
+    )
+
+    # Determine what edges need to be simplified (either between diff.
+    #  clusters or self-loops in a cluster):
+    edges_tosimplify_gdf = edges_gdf.query(
+        "origin_cluster != destination_cluster or "
+        f"(('{edge_to_att}' == '{edge_from_att}') and origin_cluster >= 0)"
+    )
+
+    # Determine the new point geometries (when exists):
+    edges_tosimplify_gdf = edges_tosimplify_gdf.assign(
+        new_origin_pt=edges_tosimplify_gdf.origin_cluster.map(
+            cluster_gdf.geometry, None
+        )
+    )
+    edges_tosimplify_gdf = edges_tosimplify_gdf.assign(
+        new_destination_pt=edges_tosimplify_gdf.destination_cluster.map(
+            cluster_gdf.geometry, None
+        )
+    )
+
+    # Determine the new geometry according to the simplification method:
+    if method == "extend":
+        edges_simplified_geometries = edges_tosimplify_gdf.apply(
+            lambda edge: _extension_simplification(
+                edge.geometry, edge.new_origin_pt, edge.new_destination_pt
+            ),
+            axis=1,
+        )
+        edges_simplified_gdf = edges_tosimplify_gdf.assign(
+            new_geometry=edges_simplified_geometries
+        )
+    elif method == "euclidean":
+        edges_simplified_geometries = edges_tosimplify_gdf.apply(
+            lambda edge: _euclidean_simplification(
+                edge.geometry, edge.new_origin_pt, edge.new_destination_pt
+            ),
+            axis=1,
+        )
+        edges_simplified_gdf = edges_tosimplify_gdf.assign(
+            new_geometry=edges_simplified_geometries
+        )
+    elif method == "spider":
+        edges_simplified_geometries = edges_tosimplify_gdf.apply(
+            lambda edge: _spider_simplification(
+                edge.geometry, edge.new_origin_pt, edge.new_destination_pt, buffer
+            ),
+            axis=1,
+        )
+        edges_simplified_gdf = edges_tosimplify_gdf.assign(
+            new_geometry=edges_simplified_geometries
+        )
+    else:
+        msg = (
+            f"Simplification '{method}' not recognized. See documentation for options."
+        )
+        raise ValueError(msg)
+
+    # Rename and update the columns:
+    cols_rename = {
+        edge_from_att: "original_from",
+        edge_to_att: "original_to",
+        "origin_cluster": edge_from_att,
+        "destination_cluster": edge_to_att,
+        "geometry": "original_geometry",
+    }
+    new_edges_gdf = edges_simplified_gdf.rename(cols_rename, axis=1)
+
+    cols_drop = ["new_origin_pt", "new_destination_pt"]
+    new_edges_gdf = new_edges_gdf.drop(columns=cols_drop)
+
+    new_edges_gdf = new_edges_gdf.set_geometry("new_geometry")
+    new_edges_gdf.loc[:, "length"] = new_edges_gdf.length
+
+    # Update the indices:
+    new_edges_gdf.loc[:, edge_from_att] = new_edges_gdf[edge_from_att].where(
+        new_edges_gdf[edge_from_att] >= 0, new_edges_gdf["original_from"]
+    )
+    new_edges_gdf.loc[:, edge_to_att] = new_edges_gdf[edge_to_att].where(
+        new_edges_gdf[edge_to_att] >= 0, new_edges_gdf["original_to"]
+    )
+
+    # Get the edge list with (from, to, data):
+    new_edges_list = list(
+        zip(
+            new_edges_gdf[edge_from_att],
+            new_edges_gdf[edge_to_att],
+            new_edges_gdf.iloc[:, 2:].to_dict("index").values(),
+            strict=False,
+        )
+    )
+
+    return new_edges_list
+
+
+def _extension_simplification(geometry, new_origin, new_destination):
+    """
+    Extends edge geometry to new endpoints.
+
+    If either new_origin or new_destination is None, maintains the
+      respective current endpoint.
+
+    Parameters
+    ----------
+    geometry : shapely.LineString
+    new_origin : shapely.Point or None
+    new_destination: shapely.Point or None
+
+    Returns
+    ----------
+    shapely.LineString
+
+    """
+    # If we are dealing with a self-loop the line has no endpoints:
+    if new_origin == new_destination:
+        current_node = Point(geometry.coords[0])
+        geometry = linemerge([LineString([new_origin, current_node]), geometry])
+    # Assuming the line is not closed, we can find its endpoints:
+    else:
+        current_origin, current_destination = geometry.boundary.geoms
+        if new_origin is not None:
+            geometry = linemerge([LineString([new_origin, current_origin]), geometry])
+        if new_destination is not None:
+            geometry = linemerge(
+                [geometry, LineString([current_destination, new_destination])]
+            )
+    return geometry
+
+
+def _spider_simplification(geometry, new_origin, new_destination, buff=15):
+    """
+    Extends edge geometry to new endpoints via a "spider-web" method. Breaks
+      current geometry within a buffer of the new endpoint and then extends
+      it linearly. Useful to maintain planarity.
+
+    If either new_origin or new_destination is None, maintains the
+      respective current endpoint.
+
+    Parameters
+    ----------
+    geometry : shapely.LineString
+    new_origin : shapely.Point or None
+    new_destination: shapely.Point or None
+    buff : float
+        distance from new endpoint to break current geometry
+
+    Returns
+    ----------
+    shapely.LineString
+
+    """
+    # If we are dealing with a self-loop the line has no boundary
+    # . and we just use the first coordinate:
+    if new_origin == new_destination:
+        current_node = Point(geometry.coords[0])
+        geometry = linemerge([LineString([new_origin, current_node]), geometry])
+    # Assuming the line is not closed, we can find its endpoints
+    #  via the boundary attribute:
+    else:
+        current_origin, current_destination = geometry.boundary.geoms
+        if new_origin is not None:
+            # Create a buffer around the new origin:
+            new_origin_buffer = new_origin.buffer(buff)
+            # Use shapely.ops.split to break the edge where it
+            #  intersects the buffer:
+            geometry_split_by_buffer_list = list(
+                split(geometry, new_origin_buffer).geoms
+            )
+            # If only one geometry results, edge does not intersect
+            #  buffer and line should connect new origin to old origin
+            if len(geometry_split_by_buffer_list) == 1:
+                geometry_split_by_buffer = geometry_split_by_buffer_list[0]
+                splitting_point = current_origin
+            # If more than one geometry, merge all linestrings
+            #  but the first and get their origin
+            else:
+                geometry_split_by_buffer = linemerge(geometry_split_by_buffer_list[1:])
+                splitting_point = geometry_split_by_buffer.boundary.geoms[0]
+            # Merge this into new geometry:
+            additional_line = [LineString([new_origin, splitting_point])]
+            # Consider MultiLineStrings separately:
+            if geometry_split_by_buffer.geom_type == "MultiLineString":
+                geometry = linemerge(
+                    additional_line + list(geometry_split_by_buffer.geoms)
+                )
+            else:
+                geometry = linemerge(additional_line + [geometry_split_by_buffer])
+
+        if new_destination is not None:
+            # Create a buffer around the new destination:
+            new_destination_buffer = new_destination.buffer(buff)
+            # Use shapely.ops.split to break the edge where it
+            #  intersects the buffer:
+            geometry_split_by_buffer_list = list(
+                split(geometry, new_destination_buffer).geoms
+            )
+            # If only one geometry results, edge does not intersect
+            # . buffer and line should connect new destination to old destination
+            if len(geometry_split_by_buffer_list) == 1:
+                geometry_split_by_buffer = geometry_split_by_buffer_list[0]
+                splitting_point = current_destination
+            # If more than one geometry, merge all linestrings
+            #  but the last and get their destination
+            else:
+                geometry_split_by_buffer = linemerge(geometry_split_by_buffer_list[:-1])
+                splitting_point = geometry_split_by_buffer.boundary.geoms[1]
+            # Merge this into new geometry:
+            additional_line = [LineString([splitting_point, new_destination])]
+            # Consider MultiLineStrings separately:
+            if geometry_split_by_buffer.geom_type == "MultiLineString":
+                geometry = linemerge(
+                    list(geometry_split_by_buffer.geoms) + additional_line
+                )
+            else:
+                geometry = linemerge([geometry_split_by_buffer] + additional_line)
+
+    return geometry
+
+
+def _euclidean_simplification(geometry, new_origin, new_destination):
+    """
+    Rebuilds edge geometry to new endpoints. Ignores current geometry
+      and traces a straight line between new endpoints.
+
+    If either new_origin or new_destination is None, maintains the
+      respective current endpoint.
+
+    Parameters
+    ----------
+    geometry : shapely.LineString
+    new_origin : shapely.Point or None
+    new_destination : shapely.Point or None
+
+    Returns
+    ----------
+    shapely.LineString
+
+    """
+    # If we are dealing with a self-loop, geometry will be null!
+    if new_origin == new_destination:
+        geometry = None
+    # Assuming the line is not closed, we can find its endpoints:
+    else:
+        current_origin, current_destination = geometry.boundary.geoms
+        if new_origin is not None:
+            if new_destination is not None:
+                geometry = LineString([new_origin, new_destination])
+            else:
+                geometry = LineString([new_origin, current_destination])
+        else:
+            if new_destination is not None:
+                geometry = LineString([current_origin, new_destination])
+    return geometry
+
+
 class FaceArtifacts:
-    """Identify face artifacts in street networks
+    """Identify face artifacts in street networks.
 
     For a given street network composed of transportation-oriented geometry containing
     features representing things like roundabouts, dual carriegaways and complex
@@ -1061,7 +1543,7 @@ class FaceArtifacts:
         ['circlular_compactness', 'isoperimetric_quotient', 'diameter_ratio'], by
         default "circular_compactness"
     height_mins : float, optional
-        Required depth of valleys, by default np.NINF
+        Required depth of valleys, by default -np.inf
     height_maxs : float, optional
         Required height of peaks, by default 0.008
     prominence : float, optional
@@ -1097,6 +1579,12 @@ class FaceArtifacts:
     10  POLYGON ((-744101.275 -1043738.053, -744103.80...             2.862871
     12  POLYGON ((-744095.511 -1043623.478, -744095.35...             3.712403
     17  POLYGON ((-744488.466 -1044533.317, -744489.33...             5.158554
+
+    Notes
+    -----
+    In purely synthetic scenarios where all calculated face artifact index values are
+    equal (e.g. a regular grid/lattice) a ``LinAlgError`` error will be raised. This
+    is virtually guaranteed to not happen in practice.
     """
 
     def __init__(
@@ -1118,15 +1606,30 @@ class FaceArtifacts:
         # Polygonize street network
         polygons = gpd.GeoSeries(
             shapely.polygonize(  # polygonize
-                [gdf.unary_union]
-            ).geoms,  # get parts of the collection from polygonize
-            crs=gdf.crs,
-        ).explode(
-            ignore_index=True
-        )  # shoudln't be needed but doesn't hurt to ensure
+                [gdf.dissolve().geometry.item()]
+            )
+        ).explode(ignore_index=True)
 
         # Store geometries as a GeoDataFrame
         self.polygons = gpd.GeoDataFrame(geometry=polygons)
+
+        if self.polygons.empty:
+            warnings.warn(
+                "Input roads could not not be polygonized. "
+                "Identification of face artifacts not possible.",
+                UserWarning,
+                stacklevel=2,
+            )
+            self.kde = None
+            self.pdf = None
+            self.peaks = None
+            self.d_peaks = None
+            self.valleys = None
+            self.d_valleys = None
+            self.threshold = None
+            self.face_artifacts = None
+            return
+
         if index == "circular_compactness":
             self.polygons["face_artifact_index"] = np.log(
                 shape.minimum_bounding_circle_ratio(polygons) * polygons.area
