@@ -2,7 +2,7 @@ import geopandas as gpd
 import numpy as np
 import pytest
 from geopandas.testing import assert_geodataframe_equal
-from shapely import affinity
+from shapely import affinity, force_3d
 from shapely.geometry import LineString, MultiPoint, Point, Polygon
 from shapely.ops import polygonize
 
@@ -67,6 +67,16 @@ class TestPreprocessing:
         df = self.df_streets.drop([4, 7, 17, 22])
         assert_geodataframe_equal(df, mm.remove_false_nodes(df))
 
+        # check 3d coords in loop
+        line_1 = LineString((Point(1, 1), Point(2, 2)))
+        line_2 = LineString((Point(2, 2), Point(3, 3)))
+        line_3 = force_3d(LineString((Point(2, 2), Point(1, 2))))
+        line_4 = LineString((Point(1, 2), Point(1, 3)))
+        line_5 = LineString((Point(1, 3), Point(2, 2)))
+        with_3d = np.array([line_1, line_2, line_3, line_4, line_5])
+        no_3d = mm.remove_false_nodes(with_3d)
+        assert len(no_3d) == 3
+
     def test_CheckTessellationInput(self):
         df = self.df_buildings
         df.loc[144, "geometry"] = Polygon([(0, 0), (0, 1), (1, 0)])
@@ -256,3 +266,15 @@ def test_FaceArtifacts():
 
     with pytest.raises(ValueError, match="'banana' is not supported"):
         mm.FaceArtifacts(gdf, index="banana")
+
+    p1, p2, p3, p4 = Point(1, 0), Point(2, 0), Point(3, 0), Point(2, 1)
+    inverted_t = [LineString((p1, p2)), LineString((p2, p3)), LineString((p2, p4))]
+
+    with pytest.warns(
+        UserWarning,
+        match=(
+            "Input roads could not not be polygonized. "
+            "Identification of face artifacts not possible."
+        ),
+    ):
+        mm.FaceArtifacts(gpd.GeoDataFrame(geometry=inverted_t))
